@@ -20,14 +20,17 @@ from ctypes import (
     create_unicode_buffer
 )
 
-from typing import Callable, List
+from typing import Callable, List, Tuple
 
 from JABWrapper.jab_types import (
     AccessBridgeVersionInfo,
     AccessibleActions,
     AccessibleActionsToDo,
     AccessibleKeyBindings,
+    AccessibleRelationSetInfo,
+    AccessibleTextAttributesInfo,
     AccessibleTextItemsInfo,
+    AccessibleTextRectInfo,
     JavaObject,
     AccessibleContextInfo,
     AccessibleTextInfo,
@@ -145,27 +148,88 @@ class JavaAccessBridgeWrapper:
         self._remove_callbacks()
 
     def _define_functions(self) -> None:
-        # BOOL isJavaWindow(HWND window)
-        self._wab.isJavaWindow.argtypes = [wintypes.HWND]
-        self._wab.isJavaWindow.restype = wintypes.BOOL
         # void Windows_run()
         self._wab.Windows_run.argtypes = []
         self._wab.Windows_run.restype = None
-        # BOOL GetAccessibleContextFromHWND(HWND window, *vmID, *context)
-        self._wab.getAccessibleContextFromHWND.argtypes = [wintypes.HWND, POINTER(c_long), POINTER(JavaObject)]
-        self._wab.getAccessibleContextFromHWND.restype = wintypes.BOOL
+
         # void ReleaseJavaObject(long vmID, AccessibleContext context)
         self._wab.releaseJavaObject.argtypes = [c_long, JavaObject]
         self._wab.releaseJavaObject.restype = None
+
+        # BOOL GetVersionInfo(long vmID, AccessBridgeVersionInfo *info)
+        self._wab.getVersionInfo.argtypes = [c_long, POINTER(AccessBridgeVersionInfo)]
+        self._wab.getVersionInfo.restype = wintypes.BOOL
+
+        # Accessible context
+        # BOOL isJavaWindow(HWND window)
+        self._wab.isJavaWindow.argtypes = [wintypes.HWND]
+        self._wab.isJavaWindow.restype = wintypes.BOOL
+        # BOOL isSameObject(long vmID, AccessibleContext context_from, AccessibleContext context_to)
+        self._wab.isSameObject.argtypes = [c_long, JavaObject, JavaObject]
+        self._wab.isSameObject.restypes = wintypes.BOOL
+        # BOOL GetAccessibleContextFromHWND(HWND window, long *vmID, AccessibleContext *context)
+        self._wab.getAccessibleContextFromHWND.argtypes = [wintypes.HWND, POINTER(c_long), POINTER(JavaObject)]
+        self._wab.getAccessibleContextFromHWND.restype = wintypes.BOOL
+        # HWND getHWNDFromAccessibleContext(long vmID, AccessibleContext context)
+        self._wab.getHWNDFromAccessibleContext.argtypes = [c_long, JavaObject]
+        self._wab.getHWNDFromAccessibleContext.restype = wintypes.HWND
+        # BOOL getAccessibleContextAt(long vmID, AccessibleContext parent, int x, int y, AccessibleContext *context)
+        self._wab.getAccessibleContextAt.argtypes = [c_long, JavaObject, c_int, c_int, POINTER(JavaObject)]
+        self._wab.getAccessibleContextAt.restype = wintypes.BOOL
+        # TODO: getAccessibleContextWithFocus
         # BOOL getAccessibleContextInfo(long vmID, AccessibleContext context, AccessibleContextInfo *info)
         self._wab.getAccessibleContextInfo.argtypes = [c_long, JavaObject, POINTER(AccessibleContextInfo)]
         self._wab.getAccessibleContextInfo.restype = wintypes.BOOL
         # AccessibleContext getAccessibleChildFromContext(long vmID, AccessibleContext context, int integer)
         self._wab.getAccessibleChildFromContext.argtypes = [c_long, JavaObject, c_int]
         self._wab.getAccessibleChildFromContext.restype = JavaObject
-        # BOOL GetVersionInfo(long vmID, AccessBridgeVersionInfo *info)
-        self._wab.getVersionInfo.argtypes = [c_long, POINTER(AccessBridgeVersionInfo)]
-        self._wab.getVersionInfo.restype = wintypes.BOOL
+        # JavaObject getAccessibleParentFromContext(c_long vmID, JavaObject child_context)
+        self._wab.getAccessibleParentFromContext.argtypes = [c_long, JavaObject]
+        self._wab.getAccessibleParentFromContext.restype = JavaObject
+
+        # Accessible table
+        # TODO: getAccessibleTableInfo
+        # TODO: getAccessibleTableCellInfo
+        # TODO: getAccessibleTableRowHeader
+        # TODO: getAccessibleTableColumnHeader
+        # TODO: getAccessibleTableRowDescription
+        # TODO: getAccessibleTableColumnDescription
+        # TODO: getAccessibleTableRowSelectionCount
+        # TODO: isAccessibleTableRowSelected
+        # TODO: getAccessibleTableRowSelections
+        # TODO: getAccessibleTableColumnSelectionCount
+        # TODO: isAccessibleTableColumnSelected
+        # TODO: getAccessibleTableColumnSelections
+        # TODO: getAccessibleTableRow
+        # TODO: getAccessibleTableColumn
+        # TODO: getAccessibleTableIndex
+
+        # AccessibleRelationSet
+        # BOOL getAccessibleRelationSet(long vmID, AccessibleContext accessibleContext, AccessibleRelationSetInfo *relationSetInfo)
+        self._wab.getAccessibleRelationSet.argtypes = [c_long, JavaObject, POINTER(AccessibleRelationSetInfo)]
+        self._wab.getAccessibleRelationSet.restype = wintypes.BOOL
+
+        # AccessibleHypertext
+        # TODO: getAccessibleHypertext
+        # TODO: activateAccessibleHyperlink
+        # TODO: getAccessibleHyperlinkCount
+        # TODO: getAccessibleHypertextExt
+        # TODO: getAccessibleHypertextLinkIndex
+        # TODO: getAccessibleHyperlink
+
+        # Accessible KeyBindings, Icons and Actions
+        # BOOL getAccessibleKeyBindings(long vmID, AccessibleContext context, AccessibleKeyBindings *bindings)
+        self._wab.getAccessibleKeyBindings.argtypes = [c_long, JavaObject, POINTER(AccessibleKeyBindings)]
+        self._wab.getAccessibleKeyBindings.restypes = wintypes.BOOL
+        # TODO: getAccessibleIcons
+        # BOOL getAccessibleActions(long vmID, AccessibleContext context, AccessibleActions *actions)
+        self._wab.getAccessibleActions.argtypes = [c_long, JavaObject, POINTER(AccessibleActions)]
+        self._wab.getAccessibleActions.restypes = wintypes.BOOL
+        # BOOL doAccessibleActions(long vmID, AccessibleContext context, AccessibleActionsToDo actionsToDo, bool *result, int *failure_index)
+        self._wab.doAccessibleActions.argtypes = [c_long, JavaObject, AccessibleActionsToDo, POINTER(c_int)]
+        self._wab.doAccessibleActions.restypes = wintypes.BOOL
+
+        # AccessibleText
         # AccessibleTextInfo GetAccessibleTextInfo(long vmID, AccessibleContext context, AccessibleTextInfo *info, int x, int y)
         self._wab.getAccessibleTextInfo.argtypes = [c_long, JavaObject, POINTER(AccessibleTextInfo), c_int, c_int]
         self._wab.getAccessibleTextInfo.restype = wintypes.BOOL
@@ -175,27 +239,61 @@ class JavaAccessBridgeWrapper:
         # BOOL GetAccessibleTextSelectionInfo(long vmID, AccessibleContext context, AccessibleTextSelectionInfo *textSelection)
         self._wab.getAccessibleTextSelectionInfo.argtypes = [c_long, JavaObject, POINTER(AccessibleTextSelectionInfo)]
         self._wab.getAccessibleTextSelectionInfo.restype = wintypes.BOOL
-        # BOOL GetCurrentAccessibleValueFromContext(long vmID, AccessibleContext context, wintypes.WCHAR *value, short len)
-        self._wab.getCurrentAccessibleValueFromContext.argtypes = [c_long, JavaObject, POINTER(wintypes.WCHAR), c_short]
+        # BOOL getAccessibleTextAttributes(long vmID, AccessibleContext context, int index, AccessibleTextAttributesInfo *attributesInfo)
+        self._wab.getAccessibleTextAttributes.argtypes = [c_long, JavaObject, c_int, POINTER(AccessibleTextAttributesInfo)]
+        self._wab.getAccessibleTextAttributes.restype = wintypes.BOOL
+        # BOOL getAccessibleTextRect(long vmID, AccessibleContext context, AccessibleTextRectInfo *rectInfo, int index)
+        self._wab.getAccessibleTextRect.argtypes = [c_long, JavaObject, POINTER(AccessibleTextRectInfo), c_int]
+        self._wab.getAccessibleTextRect.restype = wintypes.BOOL
+        # BOOL getAccessibleTextLineBounds(long vmID, AccessibleContext context, int index, int *startIndex, int *endIndex)
+        self._wab.getAccessibleTextLineBounds.argtypes = [c_long, JavaObject, c_int, POINTER(c_int), POINTER(c_int)]
+        self._wab.getAccessibleTextLineBounds.restype = wintypes.BOOL
+        # BOOL getAccessibleTextRange(long vmID, AccessibleContext context, int start, int end, c_wchar_p *text, short len)
+        self._wab.getAccessibleTextRange.argtypes = [c_long, JavaObject, c_int, c_int, c_wchar_p, c_short]
+        self._wab.getAccessibleTextRange.restype = wintypes.BOOL
+
+        # AccessibleValue
+        # BOOL getCurrentAccessibleValueFromContext(long vmID, AccessibleContext context, wintypes.WCHAR *value, short len)
+        self._wab.getCurrentAccessibleValueFromContext.argtypes = [c_long, JavaObject, c_wchar_p, c_short]
         self._wab.getCurrentAccessibleValueFromContext.restype = wintypes.BOOL
-        # BOOL getAccessibleActions(long vmID, AccessibleContext context, AccessibleActions *actions)
-        self._wab.getAccessibleActions.argtypes = [c_long, JavaObject, POINTER(AccessibleActions)]
-        self._wab.getAccessibleActions.restypes = wintypes.BOOL
-        # BOOL doAccessibleActions(long vmID, AccessibleContext context, AccessibleActionsToDo actionsToDo, bool *result, int *failure_index)
-        self._wab.doAccessibleActions.argtypes = [c_long, JavaObject, AccessibleActionsToDo, POINTER(c_int)]
-        self._wab.doAccessibleActions.restypes = wintypes.BOOL
+        # BOOL getMaximumAccessibleValueFromContext(long vmID, AccessibleContext context, wintypes.WCHAR *value, short len)
+        self._wab.getMaximumAccessibleValueFromContext.argtypes = [c_long, JavaObject, c_wchar_p, c_short]
+        self._wab.getMaximumAccessibleValueFromContext.restype = wintypes.BOOL
+        # BOOL getMinimumAccessibleValueFromContext(long vmID, AccessibleContext context, wintypes.WCHAR *value, short len)
+        self._wab.getMinimumAccessibleValueFromContext.argtypes = [c_long, JavaObject, c_wchar_p, c_short]
+        self._wab.getMinimumAccessibleValueFromContext.restype = wintypes.BOOL
+
+        # AccessibleSelection
+        # TODO: addAccessibleSelectionFromContext
+        # TODO: clearAccessibleSelectionFromContext
+        # TODO: getAccessibleSelectionFromContext
+        # TODO: getAccessibleSelectionCountFromContext
+        # TODO: isAccessibleChildSelectedFromContext
+        # TODO: removeAccessibleSelectionFromContext
+        # TODO: selectAllAccessibleSelectionFromContext
+
+        # Utility
         # BOOL setTextContents(long vmID, AccessibleContext context, str text)
         self._wab.setTextContents.argtypes = [c_long, JavaObject, wintypes.WCHAR * MAX_STRING_SIZE]
         self._wab.setTextContents.restypes = wintypes.BOOL
+        # TODO: getParentWithRole
+        # TODO: getParentWithRoleElseRoot
+        # TODO: getTopLevelObject
+        # TODO: getObjectDepth
+        # TODO: getActiveDescendent
+        # BOOL getVirtualAccessibleNameFP(long vmID, AccessibleContext context, str name, int len)
+        self._wab.getVirtualAccessibleName.argtypes = [c_long, JavaObject, wintypes.WCHAR * MAX_STRING_SIZE, c_int]
+        self._wab.getVirtualAccessibleName.restype = wintypes.BOOL
         # BOOL requestFocus(long vmID, AccessibleContext context)
         self._wab.requestFocus.argtypes = [c_long, JavaObject]
         self._wab.requestFocus.restypes = wintypes.BOOL
-        # BOOL getAccessibleKeyBindings(long vmID, AccessibleContext context, AccessibleKeyBindings *bindings)
-        self._wab.getAccessibleKeyBindings.argtypes = [c_long, JavaObject, POINTER(AccessibleKeyBindings)]
-        self._wab.getAccessibleKeyBindings.restypes = wintypes.BOOL
-        # BOOL isSameObject(long vmID, AccessibleContext context1, AccessibleContext context2)
-        self._wab.isSameObject.argtypes = [c_long, JavaObject, JavaObject]
-        self._wab.isSameObject.restypes = wintypes.BOOL
+        # TODO: selectTextRange
+        # TODO: getTextAttributesInRange
+        # TODO: getVisibleChildrenCount
+        # TODO: getVisibleChildren
+        # TODO: setCaretPosition
+        # TODO: getCaretLocation
+        # TODO: getEventsWaitingFP
 
     def _define_callbacks(self) -> None:
         # Property events
@@ -358,6 +456,13 @@ class JavaAccessBridgeWrapper:
 
         return True
 
+    def set_hwnd(self, hwnd: wintypes.HWND) -> None:
+        self._hwnd = hwnd
+
+    def set_context(self, vm_id: c_long, context: JavaObject) -> None:
+        self._vmID = vm_id
+        self.context = context
+
     def get_current_windows_handle(self) -> wintypes.HWND:
         return self._hwnd
 
@@ -386,9 +491,39 @@ class JavaAccessBridgeWrapper:
             self.context,
         ))
 
+    def get_accessible_context_from_hwnd(self, hwnd: wintypes.HWND) -> Tuple[c_long, JavaObject]:
+        vm_id = c_long()
+        context = JavaObject()
+        ok = self._wab.getAccessibleContextFromHWND(hwnd, byref(vm_id), byref(context))
+        if not ok:
+            raise APIException("Failed to get accessible context from HWND")
+        return vm_id, context
+
+    def get_hwnd_from_accessible_context(self, context) -> wintypes.HWND:
+        return self._wab.getHWNDFromAccessibleContext(self._vmID, context)
+
+    def get_accessible_context_at(self, parent: JavaObject, x: int, y: int) -> JavaObject:
+        context = JavaObject()
+        ok = self._wab.getAccessibleContextAt(self._vmID, parent, x, y, byref(context))
+        if not ok:
+            raise APIException("Failed to get accessible context at={x},{y}")
+        return context
+
     def get_child_context(self, context: JavaObject, index: c_int, ) -> JavaObject:
         return self._wab.getAccessibleChildFromContext(self._vmID, context, index)
 
+    def get_accessible_parent_from_context(self, context) -> JavaObject:
+        return self._wab.getAccessibleParentFromContext(self._vmID, context)
+
+    def get_accessible_relation_set_info(self, context: JavaObject) -> AccessibleRelationSetInfo:
+        relation_set_info = AccessibleRelationSetInfo()
+        logging.info("getting rel set")
+        ok = self._wab.getAccessibleRelationSet(self._vmID, context, byref(relation_set_info))
+        logging.info(f"rel set={ok}")
+        if not ok:
+            raise APIException("Failed to get accessible relation set info")
+        return relation_set_info
+
     def get_context_info(self, context: JavaObject) -> AccessibleContextInfo:
         info = AccessibleContextInfo()
         ok = self._wab.getAccessibleContextInfo(self._vmID, context, byref(info))
@@ -424,13 +559,56 @@ class JavaAccessBridgeWrapper:
             raise APIException("Failed to get accessible text selection info")
         return info
 
+    def get_accessible_text_attributes(self, context: JavaObject, index: int) -> AccessibleTextAttributesInfo:
+        attributes_info = AccessibleTextAttributesInfo()
+        ok = self._wab.getAccessibleTextAttributes(self._vmID, context, index, attributes_info)
+        if not ok:
+            raise APIException("Failed to get accessible text attributes info")
+        return attributes_info
+
+    def get_accessible_text_rect(self, context: JavaObject, index: int) -> AccessibleTextRectInfo:
+        rect_info = AccessibleTextRectInfo()
+        ok = self._wab.getAccessibleTextRect(self._vmID, context, byref(rect_info), index)
+        if not ok:
+            raise APIException("Failed to get accessible text rect info")
+        return rect_info
+
+    def get_accessible_text_line_bounds(self, context, index) -> Tuple[int, int]:
+        start_index = c_int()
+        end_index = c_int()
+        ok = self._wab.getAccessibleTextLineBounds(self._vmID, context, index, byref(start_index), byref(end_index))
+        if not ok:
+            raise APIException(f"Failed to get accessible text line bounds at={index}")
+        return start_index, end_index
+
+    def get_accessible_text_range(self, context: JavaObject, start_index: c_int, end_index: c_int, length: c_short) -> str:
+        buf = create_unicode_buffer(length)
+        ok = self._wab.getAccessibleTextRange(self._vmID, context, start_index, end_index, buf, length)
+        if not ok:
+            raise APIException("Failed to get accessible range")
+        return buf.value
+
     def get_current_accessible_value_from_context(self, context: JavaObject) -> str:
-        buf = create_unicode_buffer(SHORT_STRING_SIZE + 1)
+        buf = create_unicode_buffer(SHORT_STRING_SIZE)
         ok = self._wab.getCurrentAccessibleValueFromContext(self._vmID, context, buf, SHORT_STRING_SIZE)
         if not ok:
             raise APIException("Failed to get current accessible value from context")
         return buf.value
 
+    def get_maximum_accessible_value_from_context(self, context: JavaObject) -> str:
+        buf = create_unicode_buffer(SHORT_STRING_SIZE)
+        ok = self._wab.getMaximumAccessibleValueFromContext(self._vmID, context, buf, SHORT_STRING_SIZE)
+        if not ok:
+            raise APIException("Failed to get maximum accessible value from context")
+        return buf.value
+
+    def get_minimum_accessible_value_from_context(self, context: JavaObject) -> str:
+        buf = create_unicode_buffer(SHORT_STRING_SIZE)
+        ok = self._wab.getMinimumAccessibleValueFromContext(self._vmID, context, buf, SHORT_STRING_SIZE)
+        if not ok:
+            raise APIException("Failed to get minimum accessible value from context")
+        return buf.value
+
     def get_accessible_actions(self, context: JavaObject) -> AccessibleActions:
         actions = AccessibleActions()
         ok = self._wab.getAccessibleActions(self._vmID, context, byref(actions))
@@ -465,6 +643,13 @@ class JavaAccessBridgeWrapper:
     def is_same_object(self, context_from: JavaObject, context_to: JavaObject) -> bool:
         return self._wab.isSameObject(self._vmID, context_from, context_to)
 
+    def get_virtual_accessible_name(self, context: JavaObject):
+        buf = create_unicode_buffer(MAX_STRING_SIZE)
+        ok = self._wab.getVirtualAccessibleName(self._vmID, context, buf, MAX_STRING_SIZE)
+        if not ok:
+            raise APIException("Failed to get virtual accessible name")
+        return buf.value
+
     def register_callback(self, name: str, callback: Callable[[JavaObject], None]) -> None:
         logging.debug(f"Registering callback={name}")
         if name in self._context_callbacks:
